{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE TypeOperators #-}
{- ToDo:
provide inplace updates of blocks: add, scale
-}
{- ToDo:
conversion from Block to Chunk
-}
module Numeric.BLAS.Matrix.RowMajor.Block (
   Array, Matrix, Block,
   shape, height, width,
   reshapeHeight, reshapeWidth,
   mapHeight, mapWidth,
   toMatrix, toMutableMatrix, toMutableSubmatrix,
   block, transposed, transposable, zero,
   submatrix, transposedSubmatrix,
   row, column,
   above, beside, (&===), (&|||),
   fromBlockColumn, fromBlockRow, fromBlockEnsemble,
   ) where

import qualified Numeric.BLAS.Matrix.Transposable as Transp
import qualified Numeric.BLAS.Matrix.RowMajor as RowMajor
import qualified Numeric.BLAS.Vector.SlicePrivate as VectorSlice
import qualified Numeric.BLAS.Scalar as Scalar
-- import qualified Numeric.BLAS.Subobject.View as View
import qualified Numeric.BLAS.Subobject.Shape as Subshape
import qualified Numeric.BLAS.FFI.Generic as Blas
import qualified Numeric.CBLAS.FFI.Extra as CBlas
import qualified Numeric.Netlib.Modifier as Modi
import qualified Numeric.Netlib.Utility as Call
import qualified Numeric.Netlib.Class as Class

import Foreign.Marshal.Array (advancePtr)
import Foreign.ForeignPtr (withForeignPtr)
import Foreign.Ptr (Ptr)
import Foreign.Storable (Storable)

import qualified Data.Array.Comfort.Boxed.Dim2 as BoxedArray2
import qualified Data.Array.Comfort.Boxed as BoxedArray
import qualified Data.Array.Comfort.Storable.Mutable.Private as MutArray
import qualified Data.Array.Comfort.Storable.Unchecked as Array
import qualified Data.Array.Comfort.Shape.SubSize as SubSize
import qualified Data.Array.Comfort.Shape as Shape
-- import qualified Data.List.HT as ListHT
import Data.Array.Comfort.Shape ((::+)((::+)))
import Data.Foldable (for_)

import Control.Monad.Primitive (PrimMonad, unsafeIOToPrim)
import Control.Monad.Trans.Cont (evalContT)
import Control.Monad.IO.Class (liftIO)


{- $setup
>>> :set -XTypeOperators
>>> import Test.NumberType_.Numeric.BLAS.Matrix.RowMajor (Matrix, forMatrix)
>>> import Test.NumberType_.Numeric.BLAS.Vector
>>>             (forVector, genVector, Number_, number_, maxDim)
>>> import Test.Slice (ShapeInt, shapeInt)
>>> import qualified Numeric.BLAS.Matrix.RowMajor.Block as BlockMatrix
>>> import qualified Numeric.BLAS.Matrix.RowMajor as Matrix
>>> import qualified Numeric.BLAS.Vector.Slice as VectorSlice
>>> import qualified Numeric.BLAS.Vector as Vector
>>> import qualified Numeric.BLAS.Subobject.Shape as Subshape
>>> import qualified Numeric.BLAS.Slice as Slice
>>> import Numeric.BLAS.Matrix.RowMajor.Block ((&===), (&|||))
>>> import Numeric.BLAS.Vector (Vector, (+++))
>>> import qualified Data.Array.Comfort.Boxed.Dim2 as BoxedArray2
>>> import qualified Data.Array.Comfort.Boxed as BoxedArray
>>> import qualified Data.Array.Comfort.Storable.Dim2 as Array2
>>> import qualified Data.Array.Comfort.Storable as Array
>>> import qualified Data.Array.Comfort.Shape as Shape
>>> import Data.Array.Comfort.Shape ((::+)((::+)))
>>> import Control.Monad (replicateM)
>>> import Control.Applicative (liftA2)
>>> import qualified Test.QuickCheck as QC
>>> import Test.QuickCheck ((===))
>>>
>>> forPad :: (Shape.C sh, Show sh, QC.Testable prop) =>
>>>    sh ->
>>>    (Vector (sh, ShapeInt) Number_ ->
>>>     Vector (sh, ShapeInt) Number_ ->
>>>     prop) ->
>>>    QC.Property
>>> forPad height prop =
>>>    QC.forAll (shapeInt <$> QC.choose (0,maxDim)) $ \leftPadShape ->
>>>    QC.forAll (shapeInt <$> QC.choose (0,maxDim)) $ \rightPadShape ->
>>>    QC.forAll (genVector (height, leftPadShape) number_) $ \leftPad ->
>>>    QC.forAll (genVector (height, rightPadShape) number_) $ \rightPad ->
>>>    prop leftPad rightPad
>>>
>>> forPadded :: (QC.Testable prop) =>
>>>    (Matrix Number_ ->
>>>     Vector (Subshape.Submatrix (ShapeInt, ShapeInt)) Number_ ->
>>>     prop) ->
>>>    QC.Property
>>> forPadded prop =
>>>    forMatrix number_ $ \a ->
>>>    let height = Matrix.height a in
>>>    forVector number_ $ \prefix ->
>>>    forVector number_ $ \suffix ->
>>>    forPad height $ \leftPad rightPad ->
>>>    prop a $ Array.mapShape Subshape.submatrixFromAppend $
>>>       prefix +++ Matrix.beside leftPad (Matrix.beside a rightPad) +++ suffix
-}


{-
This is quite a duplicate of the definition in Data.Array.Comfort.Storable.Dim2.
Instead of a custom definition we could define a generic GADT in comfort-array
and provide two actual construction functions,
one in comfort-array and one here.

Pros: We could reuse the (&===) and (&|||) infix operators.

Cons: We have 'transposed'
which can be efficiently implemented in comfort-blas but not in comfort-array.
-}
data Array shape a = Array shape (Int -> Ptr a -> IO ())
type Matrix height width = Array (height, width)


height :: (Block block) => block (height,width) a -> height
height = fst . shape

width :: (Block block) => block (height,width) a -> width
width = snd . shape

-- ToDo: 'assemble' might be better named 'write' or 'writeTo'
reshapeHeight ::
   (Shape.C height0, Shape.C height1) =>
   height1 -> Matrix height0 width a -> Matrix height1 width a
reshapeHeight height1 (Array (height0, width_) assemble) =
   if Shape.size height0 == Shape.size height1
      then Array (height1, width_) assemble
      else error "Block.reshapeHeight: shape sizes mismatch"

reshapeWidth ::
   (Shape.C width0, Shape.C width1) =>
   width1 -> Matrix height width0 a -> Matrix height width1 a
reshapeWidth width1 (Array (height_, width0) assemble) =
   if Shape.size width0 == Shape.size width1
      then Array (height_, width1) assemble
      else error "Block.reshapeWidth: shape sizes mismatch"

mapHeight ::
   (Shape.C height0, Shape.C height1) =>
   (height0 -> height1) -> Matrix height0 width a -> Matrix height1 width a
mapHeight f arr = reshapeHeight (f $ height arr) arr

mapWidth ::
   (Shape.C width0, Shape.C width1) =>
   (width0 -> width1) -> Matrix height width0 a -> Matrix height width1 a
mapWidth f arr = reshapeWidth (f $ width arr) arr


block ::
   (Block block, Shape.C height, Shape.C width, Class.Floating a) =>
   block (height, width) a -> Matrix height width a
block = blockPrivate

class Block block where
   shape :: block shape a -> shape
   blockPrivate ::
      (Shape.C height, Shape.C width, Class.Floating a) =>
      block (height, width) a -> Matrix height width a

instance Block Array where
   shape (Array sh _) = sh
   blockPrivate = id

instance Block Array.Array where
   shape = Array.shape
   blockPrivate (Array.Array sh@(height_,width_) a) =
      Array sh $ \ldb bPtr ->
         withForeignPtr a $ \aPtr -> do
         let m = Shape.size height_
         let n = Shape.size width_
         CBlas.copyMatrix Modi.NonTransposed n m aPtr n bPtr ldb

{- |
prop> :{
   forMatrix number_ $ \a ->
      Matrix.transpose a
      ===
      BlockMatrix.toMatrix (BlockMatrix.transposed a)
:}
-}
transposed ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   RowMajor.Matrix height width a -> Matrix width height a
transposed (Array.Array (height_,width_) a) =
   Array (width_,height_) $ \ldb bPtr ->
   withForeignPtr a $ \aPtr -> do
      let m = Shape.size height_
      let n = Shape.size width_
      CBlas.copyMatrix Modi.Transposed n m aPtr n bPtr ldb

transposable ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Transp.Transposable height width a -> Matrix height width a
transposable (Transp.NonTransposed a) = blockPrivate a
transposable (Transp.Transposed a) = transposed a


{- |
prop> :{
   forMatrix number_ $ \a ->
   let height = Matrix.height a in
   QC.forAll (shapeInt <$> QC.choose (0,maxDim)) $ \width ->

   Matrix.beside a (Vector.zero (height,width))
   ===
   BlockMatrix.toMatrix (a &||| BlockMatrix.zero height width)
:}
-}
zero ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   height -> width -> Matrix height width a
zero height_ width_ =
   Array (height_, width_) $ \ldc cPtr ->
   evalContT $ do
      transaPtr <- Call.char 'T'
      transbPtr <- Call.char 'N'
      let m = Shape.size width_
      let n = Shape.size height_
      mPtr <- Call.cint m
      nPtr <- Call.cint n
      kPtr <- Call.cint 0
      alphaPtr <- Call.number Scalar.zero
      aPtr <- Call.number Scalar.zero
      ldaPtr <- Call.leadingDim 0
      betaPtr <- Call.number Scalar.zero
      ldcPtr <- Call.leadingDim ldc
      liftIO $
         Blas.gemm
            transaPtr transbPtr mPtr nPtr kPtr
            alphaPtr aPtr ldaPtr aPtr ldaPtr betaPtr cPtr ldcPtr


{- |
prop> :{
   forMatrix number_ $ \a ->

   a
   ===
   (Matrix.reshapeHeight (Matrix.height a) $ BlockMatrix.toMatrix $
      BlockMatrix.fromBlockColumn (Matrix.width a) $ map BlockMatrix.row $
      VectorSlice.slicesVector Slice.rows a)
:}

prop> :{
   forMatrix number_ $ \a ->

   Matrix.transpose a
   ===
   (Matrix.reshapeHeight (Matrix.width a) $ BlockMatrix.toMatrix $
      BlockMatrix.fromBlockColumn (Matrix.height a) $ map BlockMatrix.row $
      VectorSlice.slicesVector Slice.columns a)
:}
-}
row ::
   (VectorSlice.C v, Shape.C width, Class.Floating a) =>
   v width a -> Matrix () width a
row x =
   Array ((), VectorSlice.shape x) $ \_ldb bPtr -> evalContT $ do
      (nPtr, xPtr, incxPtr) <- VectorSlice.sizeSliceArg x
      incbPtr <- Call.cint 1
      liftIO $ Blas.copy nPtr xPtr incxPtr bPtr incbPtr

{- |
prop> :{
   forMatrix number_ $ \a ->

   a
   ===
   (Matrix.reshapeWidth (Matrix.width a) $ BlockMatrix.toMatrix $
      BlockMatrix.fromBlockRow (Matrix.height a) $ map BlockMatrix.column $
      VectorSlice.slicesVector Slice.columns a)
:}

prop> :{
   forMatrix number_ $ \a ->

   Matrix.transpose a
   ===
   (Matrix.reshapeWidth (Matrix.height a) $ BlockMatrix.toMatrix $
      BlockMatrix.fromBlockRow (Matrix.width a) $ map BlockMatrix.column $
      VectorSlice.slicesVector Slice.rows a)
:}
-}
column ::
   (VectorSlice.C v, Shape.C height, Class.Floating a) =>
   v height a -> Matrix height () a
column x =
   Array (VectorSlice.shape x, ()) $ \ldb bPtr -> evalContT $ do
      (nPtr, xPtr, incxPtr) <- VectorSlice.sizeSliceArg x
      incbPtr <- Call.cint ldb
      liftIO $ Blas.copy nPtr xPtr incxPtr bPtr incbPtr


{-
ToDo:
We could also have a Submatrix shape,
that exposes only height and width shape
and handles offset and leading dimension internally.
-}
{- |
Take the submatrix with dimensions @height@ and @width@ from the 'Array.Array'.

prop> :{
   forPadded $ \a padded ->

   BlockMatrix.toMatrix (BlockMatrix.submatrix padded)
   ===
   a
:}

prop> :{
   forPadded $ \a padded ->
   forPad (Matrix.height a) $ \leftPad rightPad ->
   let b = BlockMatrix.submatrix padded in

   BlockMatrix.toMatrix (leftPad &||| b &||| rightPad)
   ===
   Matrix.beside leftPad (Matrix.beside a rightPad)
:}
-}
submatrix ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Array.Array (Subshape.Submatrix (height, width)) a ->
   Matrix height width a
submatrix (Array.Array sh a) =
   let (height_,width_) = Subshape.shape sh in
   Array (height_,width_) $ \ldb bPtr ->
   withForeignPtr a $ \a0Ptr -> do
      let aPtr = advancePtr a0Ptr $ Subshape.start sh
      let lda = Subshape.submatrixLeadingDim sh
      let m = Shape.size height_
      let n = Shape.size width_
      CBlas.copyMatrix Modi.NonTransposed n m aPtr lda bPtr ldb

{- |
Take the submatrix with dimensions @height@ and @width@ from the 'Array.Array'
and prepare it for transposed insertion in a new block matrix.

prop> :{
   forPadded $ \a padded ->

   BlockMatrix.toMatrix (BlockMatrix.transposedSubmatrix padded)
   ===
   Matrix.transpose a
:}

prop> :{
   forPadded $ \a padded ->
   forPad (Matrix.width a) $ \leftPad rightPad ->
   let b = BlockMatrix.transposedSubmatrix padded in

   BlockMatrix.toMatrix (leftPad &||| b &||| rightPad)
   ===
   Matrix.beside leftPad (Matrix.beside (Matrix.transpose a) rightPad)
:}
-}
transposedSubmatrix ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Array.Array (Subshape.Submatrix (height, width)) a ->
   Matrix width height a
transposedSubmatrix (Array.Array sh a) =
   let (height_,width_) = Subshape.shape sh in
   Array (width_,height_) $ \ldb bPtr ->
   withForeignPtr a $ \a0Ptr -> do
      let aPtr = advancePtr a0Ptr $ Subshape.start sh
      let lda = Subshape.submatrixLeadingDim sh
      let m = Shape.size height_
      let n = Shape.size width_
      CBlas.copyMatrix Modi.Transposed n m aPtr lda bPtr ldb


above ::
   (Shape.C heightA, Eq width, Storable a) =>
   Matrix heightA width a -> Matrix heightB width a ->
   Matrix (heightA::+heightB) width a
above (Array (heightA,widthA) a) (Array (heightB,widthB) b) =
   Array
      (if widthA == widthB
         then (heightA::+heightB, widthA)
         else error "Matrix.RowMajor.Block.above: widths mismatch")
      (\ldc cPtr -> do
         a ldc cPtr
         b ldc (advancePtr cPtr (ldc * Shape.size heightA)))

beside ::
   (Shape.C widthA, Eq height, Storable a) =>
   Matrix height widthA a -> Matrix height widthB a ->
   Matrix height (widthA::+widthB) a
beside (Array (heightA,widthA) a) (Array (heightB,widthB) b) =
   Array
      (if heightA == heightB
         then (heightA, widthA::+widthB)
         else error "Matrix.RowMajor.Block.beside: heights mismatch")
      (\ldc cPtr -> do
         a ldc cPtr
         b ldc (advancePtr cPtr (Shape.size widthA)))

infixr 2 &===
infixr 3 &|||

(&===) ::
   (Block blockA, Block blockB) =>
   (Shape.C heightA, Shape.C heightB) =>
   (Shape.C width, Eq width) =>
   (Class.Floating a) =>
   blockA (heightA,width) a -> blockB (heightB,width) a ->
   Matrix (heightA::+heightB) width a
(&===) a b = above (block a) (block b)

(&|||) ::
   (Block blockA, Block blockB) =>
   (Shape.C height, Eq height) =>
   (Shape.C widthA, Shape.C widthB) =>
   (Class.Floating a) =>
   blockA (height,widthA) a -> blockB (height,widthB) a ->
   Matrix height (widthA::+widthB) a
(&|||) a b = beside (block a) (block b)



{- |
prop> :{
   QC.forAll (shapeInt <$> QC.choose (0,maxDim)) $ \height ->
   QC.forAll (QC.choose (0,5)) $ \n ->
   QC.forAll (replicateM n
                (flip genVector number_ . (,) height . shapeInt
                   =<< QC.choose (0,maxDim))) $ \blocks ->

   BlockMatrix.toMatrix (BlockMatrix.fromBlockRow height blocks)
   ===
   BlockMatrix.toMatrix
      (foldr
         (\a b -> BlockMatrix.mapWidth (\(shA::+shB) -> shA:shB) $ a&|||b)
         (BlockMatrix.zero height [])
         blocks)
:}
-}
fromBlockRow ::
   (Block block, Shape.C height, Eq height, Shape.C width, Class.Floating a) =>
   height -> [block (height,width) a] -> Matrix height [width] a
fromBlockRow height_ as0 =
   let as = map block as0 in
   Array
      (if all ((height_ ==) . height) as
         then (height_, map width as)
         else error "Matrix.RowMajor.Block.fromBlockRow: heights mismatch")
      (\ldc cPtr ->
         let ptrs = scanl advancePtr cPtr $ map (Shape.size . width) as in
         for_ (zip ptrs as) $ \(ptr, Array _ a) -> a ldc ptr)

{- |
prop> :{
   QC.forAll (shapeInt <$> QC.choose (0,maxDim)) $ \width ->
   QC.forAll (QC.choose (0,5)) $ \n ->
   QC.forAll (replicateM n
                (flip genVector number_ . flip (,) width . shapeInt
                   =<< QC.choose (0,maxDim))) $ \blocks ->

   Matrix.transpose
      (BlockMatrix.toMatrix (BlockMatrix.fromBlockColumn width blocks))
   ===
   BlockMatrix.toMatrix
      (BlockMatrix.fromBlockRow width $ map BlockMatrix.transposed blocks)
:}
-}
fromBlockColumn ::
   (Block block, Shape.C height, Shape.C width, Eq width, Class.Floating a) =>
   width -> [block (height,width) a] -> Matrix [height] width a
fromBlockColumn width_ as0 =
   let as = map block as0 in
   Array
      (if all ((width_ ==) . width) as
         then (map height as, width_)
         else error "Matrix.RowMajor.Block.fromBlockColumn: widths mismatch")
      (\ldc cPtr ->
         let ptrs = scanl advancePtr cPtr $
                        map ((ldc*) . Shape.size . height) as in
         for_ (zip ptrs as) $ \(ptr, Array _ a) -> a ldc ptr)

{-
Does not work so nicely, because 'horiz' is not a parameter.
A type class (AssembleEnsemble ensembleWidth width = horiz) would help,
but it would exclude e.g. horiz = [width].

fromRowEnsemble ::
   (Shape.C height, Eq height) =>
   (BoxedArray.Ensemble horiz,
    BoxedArray.EnsembleOuterShape horiz ~ ensembleWidth,
    Shape.C ensembleWidth, Eq ensembleWidth,
    BoxedArray.EnsembleInnerShape horiz ~ width, Shape.C width,  Eq width) =>
   (Storable a) =>
   height ->
   BoxedArray.Array ensembleWidth (block (height,width) a) ->
   Matrix height horiz a
-}

{- |
prop> :{
   QC.forAll (QC.choose (0,5)) $ \m ->
   QC.forAll (replicateM m (shapeInt <$> QC.choose (0,maxDim))) $ \heights ->
   QC.forAll (QC.choose (0,5)) $ \n ->
   QC.forAll (replicateM n (shapeInt <$> QC.choose (0,maxDim))) $ \widths ->
   QC.forAll (traverse (flip genVector number_) $
              BoxedArray2.cartesian
                  (BoxedArray.vectorFromList heights)
                  (BoxedArray.vectorFromList widths)) $ \blocks ->

   Matrix.transpose
      (BlockMatrix.toMatrix
         (BlockMatrix.fromBlockEnsemble heights widths blocks))
   ===
   BlockMatrix.toMatrix
      (BlockMatrix.fromBlockEnsemble widths heights $
         BoxedArray2.transpose $ fmap BlockMatrix.transposed blocks)
:}
-}
fromBlockEnsemble ::
   (Block block) =>
   (BoxedArray.Ensemble vert,
    BoxedArray.EnsembleOuterShape vert ~ ensembleHeight,
    Shape.C ensembleHeight, Eq ensembleHeight,
    BoxedArray.EnsembleInnerShape vert ~ height, Shape.C height, Eq height) =>
   (BoxedArray.Ensemble horiz,
    BoxedArray.EnsembleOuterShape horiz ~ ensembleWidth,
    Shape.C ensembleWidth, Eq ensembleWidth,
    BoxedArray.EnsembleInnerShape horiz ~ width, Shape.C width,  Eq width) =>
   (Class.Floating a) =>
   vert -> horiz ->
   BoxedArray.Array (ensembleHeight, ensembleWidth) (block (height,width) a) ->
   Matrix vert horiz a
fromBlockEnsemble vert horiz blocks =
   reshapeHeight vert $ reshapeWidth horiz $
   let widths = BoxedArray.toList $ BoxedArray.fromEnsemble horiz in
   fromBlockColumn widths $
   zipWith fromBlockRow
      (BoxedArray.toList $ BoxedArray.fromEnsemble vert)
      (map BoxedArray.toList $ BoxedArray.toList $
         BoxedArray2.toRowArray blocks)
--      (ListHT.sliceVertical (length widths) (BoxedArray.toList blocks))


{- |
prop> :{
   forMatrix number_ $ \blockA1 ->
   forMatrix number_ $ \blockB3 ->
   QC.forAll
      (liftA2
         (\char0 char1 -> Shape.Range (min char0 char1) (max char0 char1))
         (QC.choose ('a','k')) (QC.choose ('a','k'))) $
      \shapeC1 ->
   let shapeR0 = fst $ Array.shape blockA1 in
   let shapeC0 = snd $ Array.shape blockA1 in
   let shapeR1 = fst $ Array.shape blockB3 in
   let shapeC2 = snd $ Array.shape blockB3 in
   QC.forAll (genVector (shapeR0, shapeC1) number_) $ \blockA2 ->
   QC.forAll (genVector (shapeR0, shapeC2) number_) $ \blockA3 ->
   QC.forAll (genVector (shapeR1, shapeC0) number_) $ \blockB1 ->
   QC.forAll (genVector (shapeR1, shapeC1) number_) $ \blockB2 ->

   BlockMatrix.toMatrix
      (blockA1 &||| Matrix.beside blockA2 blockA3
       &===
       blockB1 &||| blockB2 &||| blockB3)
   ===
   Array2.fromBlockMatrix
      (blockA1 Array2.&||| blockA2 Array2.&||| blockA3
       Array2.&===
       blockB1 Array2.&||| blockB2 Array2.&||| blockB3)
:}

prop> :{
   QC.forAll
      (liftA2
         (\char0 char1 -> Shape.Range (min char0 char1) (max char0 char1))
         (QC.choose ('a','k')) (QC.choose ('a','k'))) $
      \shapeR0 ->
   QC.forAll
         (liftA2 Shape.Shifted (QC.choose (-10,10)) (QC.choose (0,10::Int))) $
      \shapeR1 ->
   let shapeR2 = () in
   QC.forAll (fmap Shape.ZeroBased (QC.choose (0,10::Int))) $
      \shapeC0 ->
   QC.forAll (fmap Shape.OneBased (QC.choose (0,10::Int))) $
      \shapeC1 ->
   let shapeC2 :: Shape.Enumeration Ordering
       shapeC2 = Shape.Enumeration in

   QC.forAll (genVector (shapeR0, shapeC0) number_) $ \blockA1 ->
   QC.forAll (genVector (shapeR0, shapeC1) number_) $ \blockA2 ->
   QC.forAll (genVector (shapeC2, shapeR0) number_) $ \blockA3 ->
   QC.forAll (genVector (shapeR1, shapeC0) number_) $ \blockB1 ->
   QC.forAll (genVector (shapeR1, shapeC1) number_) $ \blockB2 ->
   QC.forAll (genVector (shapeR1, shapeC2) number_) $ \blockB3 ->
   QC.forAll (genVector (shapeR2, shapeC0) number_) $ \blockC1 ->
   QC.forAll (genVector (shapeR2, shapeC1) number_) $ \blockC2 ->
   QC.forAll (genVector (shapeR2, shapeC2) number_) $ \blockC3 ->

   BlockMatrix.toMatrix
      (blockA1 &||| blockA2 &||| BlockMatrix.transposed blockA3
       &===
       blockB1 &||| blockB2 &||| blockB3
       &===
       blockC1 &||| blockC2 &||| blockC3)
   ===
   Matrix.beside
      (Matrix.above blockA1 $ Matrix.above blockB1 blockC1)
      (Matrix.above
         (Matrix.beside blockA2 $ Matrix.transpose blockA3)
         (Matrix.beside
            (Matrix.above blockB2 blockC2)
            (Matrix.above blockB3 blockC3)))
:}
-}
toMatrix ::
   (Shape.C height, Shape.C width, Class.Floating a) =>
   Matrix height width a -> RowMajor.Matrix height width a
toMatrix (Array sh assemble) =
   Array.unsafeCreateWithAutoSizes sh $
      \(SubSize.Atom _, SubSize.Atom ld) -> assemble ld

toMutableMatrix ::
   (PrimMonad m, Shape.C height, Shape.C width, Class.Floating a) =>
   Matrix height width a -> m (MutArray.Array m (height,width) a)
toMutableMatrix (Array sh assemble) =
   MutArray.unsafeCreate sh $ assemble $ Shape.size $ snd sh
{-
   MutArray.unsafeCreateWithAutoSizes sh $
      \(SubSize.Atom _, SubSize.Atom ld) -> assemble ld
-}

toMutableSubmatrix ::
   (Shape.C height, Shape.C width, Eq height, Eq width) =>
   (PrimMonad m, Class.Floating a) =>
   Matrix height width a ->
   MutArray.Array m (Subshape.Submatrix (height,width)) a ->
   m ()
toMutableSubmatrix (Array shx assemble) (MutArray.Array shy y) =
      unsafeIOToPrim $ do
   Call.assert "Block.toMutableSubmatrix: shapes mismatch" $
      shx == Subshape.shape shy
   MutArray.withArrayPtr y $ \ptr ->
      assemble
         (Subshape.submatrixLeadingDim shy)
         (advancePtr ptr $ Subshape.start shy)
