module Numeric.BLAS.Matrix.RowMajor.Mutable (
   addTensorProduct,
   addTensorProductSlice,
   addSubmatrix,
   ) where

import qualified Numeric.BLAS.Subobject.Shape as Subshape
-- import qualified Numeric.BLAS.Shape as SubShape

import qualified Numeric.BLAS.Vector.SlicePrivate as VectorSlice
-- import qualified Numeric.BLAS.Slice as Slice
-- import qualified Numeric.BLAS.Scalar as Scalar
-- import Numeric.BLAS.Vector.SlicePrivate
import Numeric.BLAS.Scalar (one)
-- import Numeric.BLAS.Private (fill)

import qualified Numeric.CBLAS.FFI.Extra as CBlas
import qualified Numeric.BLAS.FFI.Generic as Blas
-- import qualified Numeric.BLAS.FFI.Complex as BlasComplex
-- import qualified Numeric.BLAS.FFI.Real as BlasReal

import qualified Numeric.Netlib.Class as Class
import qualified Numeric.Netlib.Utility as Call
import Numeric.Netlib.Modifier (Conjugation(NonConjugated,Conjugated))

import qualified Data.Array.Comfort.Storable.Mutable.Private as MutArray
import qualified Data.Array.Comfort.Storable.Unchecked as Array
-- import qualified Data.Array.Comfort.Storable as Array
-- import qualified Data.Array.Comfort.Shape.SubSize as SubSize
import qualified Data.Array.Comfort.Shape as Shape

import qualified Foreign.Marshal.Array.Guarded as ForeignArray
import Foreign.Marshal.Array (advancePtr)
import Foreign.ForeignPtr (withForeignPtr)
import Foreign.Storable (Storable)
import Foreign.Ptr (Ptr)
import Foreign.C.Types (CInt)

import Control.Monad.Primitive (PrimMonad, unsafeIOToPrim)
import Control.Monad.Trans.Cont (ContT(ContT), evalContT)
import Control.Monad.IO.Class (liftIO)
import Control.Monad (when)
import Control.Applicative (liftA2)


{- $setup
>>> :set -XTypeOperators
>>> import Test.NumberType_.Numeric.BLAS.Matrix.RowMajor (forMatrix)
>>> import Test.NumberType_.Numeric.BLAS.Vector (genVector, number_)
>>> import qualified Numeric.BLAS.Matrix.RowMajor.Mutable as MutMatrix
>>> import qualified Numeric.BLAS.Matrix.RowMajor as Matrix
>>> import qualified Numeric.BLAS.Subobject.View.Matrix as ViewMatrix
>>> import qualified Numeric.BLAS.Subobject.Shape as Subshape
>>> import Numeric.BLAS.Vector ((.*|), (|+|))
>>> import qualified Data.Array.Comfort.Storable.Mutable as MutArray
>>> import qualified Data.Array.Comfort.Storable as Array
>>> import qualified Data.Array.Comfort.Shape as Shape
>>> import Data.Array.Comfort.Shape ((::+)((::+)))
>>> import qualified Test.QuickCheck as QC
>>> import Test.QuickCheck ((===))
-}


-- ToDo: generalize to also allow Subvector
submatrixArg :: (Storable a) =>
   MutArray.Array m (Subshape.Submatrix sh) a ->
   Call.FortranIO r (Ptr a, Ptr CInt)
submatrixArg (MutArray.Array sh a) =
   liftA2 (,)
      (fmap (flip advancePtr (Subshape.start sh)) $
         ContT $ ForeignArray.withMutablePtr a)
      (Call.leadingDim $ Subshape.submatrixLeadingDim sh)

submatrixArg_ :: (Storable a) =>
   MutArray.Array m (Subshape.Submatrix sh) a ->
   Call.FortranIO r (Ptr a, Int)
submatrixArg_ (MutArray.Array sh a) =
   liftA2 (,)
      (fmap (flip advancePtr (Subshape.start sh)) $
         ContT $ ForeignArray.withMutablePtr a)
      (return $ Subshape.submatrixLeadingDim sh)

submatrixArgImm :: (Storable a) =>
   Array.Array (Subshape.Submatrix sh) a ->
   Call.FortranIO r (Ptr a, Int)
submatrixArgImm (Array.Array sh a) =
   liftA2 (,)
      (fmap (flip advancePtr (Subshape.start sh)) $
         ContT $ withForeignPtr a)
      (return $ Subshape.submatrixLeadingDim sh)

-- ToDo: could be the basis of Matrix.RowMajor.tensorProduct (with beta=0)
-- ToDo: input vectors could be Subvector but not Slice
addTensorProduct ::
   (PrimMonad m, Class.Floating a) =>
   (VectorSlice.Subvector v, VectorSlice.Subvector w) =>
   (Shape.C height, Eq height, Shape.C width, Eq width) =>
   Either Conjugation Conjugation ->
   a -> v height a -> w width a ->
   MutArray.Array m (Subshape.Submatrix (height,width)) a ->
   m ()
addTensorProduct side alpha x y z =
   unsafeIOToPrim $ do
   let height_ = VectorSlice.shape x
   let width_  = VectorSlice.shape y
   let n = Shape.size height_
   let m = Shape.size width_
   let (heightZ,widthZ) = Subshape.shape $ MutArray.shape z
   Call.assert "addTensorProduct: heights mismatch" $ height_ == heightZ
   Call.assert "addTensorProduct: widths mismatch"  $ width_  == widthZ
   let trans conjugated =
         case conjugated of NonConjugated -> 'T'; Conjugated -> 'C'
   let ((transa,transb),(lda,ldb)) =
         case side of
            Left c -> (('N', trans c),(m,n))
            Right c -> ((trans c, 'N'),(1,1))
   when (n>0 && m>0) $ evalContT $ do
      transaPtr <- Call.char transa
      transbPtr <- Call.char transb
      mPtr <- Call.cint m
      nPtr <- Call.cint n
      kPtr <- Call.cint 1
      alphaPtr <- Call.number alpha
      aPtr <- VectorSlice.startArg y
      ldaPtr <- Call.leadingDim lda
      bPtr <- VectorSlice.startArg x
      ldbPtr <- Call.leadingDim ldb
      betaPtr <- Call.number one
      (cPtr,ldcPtr) <- submatrixArg z
      liftIO $
         Blas.gemm
            transaPtr transbPtr mPtr nPtr kPtr
            alphaPtr aPtr ldaPtr bPtr ldbPtr
            betaPtr cPtr ldcPtr

-- ToDo: unify with addTensorProduct
addTensorProductSlice ::
   (VectorSlice.C v, VectorSlice.C w, PrimMonad m, Class.Floating a) =>
   (Shape.C height, Eq height, Shape.C width, Eq width) =>
   Conjugation ->
   a -> v height a -> w width a ->
   MutArray.Array m (Subshape.Submatrix (height,width)) a ->
   m ()
addTensorProductSlice conj alpha x y z =
   unsafeIOToPrim $ do
   let height_ = VectorSlice.shape x
   let width_  = VectorSlice.shape y
   let n = Shape.size height_
   let m = Shape.size width_
   let (heightZ,widthZ) = Subshape.shape $ MutArray.shape z
   Call.assert "addTensorProductSlice: heights mismatch" $ height_ == heightZ
   Call.assert "addTensorProductslice: widths mismatch"  $ width_  == widthZ
   when (n>0 && m>0) $ evalContT $ do
      (mPtr,xPtr,incxPtr) <- VectorSlice.sizeSliceArg x
      (nPtr,yPtr,incyPtr) <- VectorSlice.sizeSliceArg y
      alphaPtr <- Call.number alpha
      (aPtr,ldaPtr) <- submatrixArg z
      let ger =
            case conj of
               NonConjugated -> Blas.geru
               Conjugated -> Blas.gerc
      liftIO $ ger nPtr mPtr alphaPtr yPtr incyPtr xPtr incxPtr aPtr ldaPtr


{- |
prop> :{
   QC.forAll number_ $ \alpha ->
   QC.forAll number_ $ \beta ->
   forMatrix number_ $ \a0 ->
   let (Shape.ZeroBased m, Shape.ZeroBased n) = Array.shape a0 in
   QC.forAll (QC.choose (0, m)) $ \mk ->
   QC.forAll (QC.choose (0, n)) $ \nk ->
   let splitHeight = Shape.ZeroBased mk ::+ Shape.ZeroBased (m-mk) in
   let splitWidth  = Shape.ZeroBased nk ::+ Shape.ZeroBased (n-nk) in
   let a = Array.reshape (splitHeight, splitWidth) a0 in
   QC.forAll (genVector (splitHeight, splitWidth) number_) $ \b ->

   alpha.*|a |+| beta.*|b
   ===
   MutArray.runST (do
      bmut <- MutArray.thaw b
      -- take quandrant in two steps in order to challenge submatrix selection
      let au = Matrix.takeTop a
      let al = Matrix.takeBottom a
      let focusA view =
            Array.mapShape (Subshape.focus view . Subshape.submatrixFromMatrix)
      let focusB view =
            MutArray.mapShape
               (Subshape.focus view . Subshape.submatrixFromMatrix)
               bmut
      let b0 = focusB (ViewMatrix.left  . ViewMatrix.top)
      let b1 = focusB (ViewMatrix.right . ViewMatrix.top)
      let b2 = focusB (ViewMatrix.left  . ViewMatrix.bottom)
      let b3 = focusB (ViewMatrix.right . ViewMatrix.bottom)
      MutMatrix.addSubmatrix alpha (focusA ViewMatrix.left  au) beta b0
      MutMatrix.addSubmatrix alpha (focusA ViewMatrix.right au) beta b1
      MutMatrix.addSubmatrix alpha (focusA ViewMatrix.left  al) beta b2
      MutMatrix.addSubmatrix alpha (focusA ViewMatrix.right al) beta b3
      return bmut)
:}
-}
addSubmatrix ::
   (PrimMonad m, Class.Floating a) =>
   (Shape.C height, Eq height, Shape.C width, Eq width) =>
   a -> Array.Array (Subshape.Submatrix (height,width)) a ->
   a -> MutArray.Array m (Subshape.Submatrix (height,width)) a ->
   m ()
addSubmatrix alpha a beta b =
   unsafeIOToPrim $ do
   let (heightA,widthA) = Subshape.shape $    Array.shape a
   let (heightB,widthB) = Subshape.shape $ MutArray.shape b
   Call.assert "addSubmatrix: heights mismatch" $ heightA == heightB
   Call.assert "addSubmatrix: widths mismatch"  $ widthA  == widthB
   let n = Shape.size heightA
   let m = Shape.size widthA
   when (n>0 && m>0) $ evalContT $ do
      (aPtr,lda) <- submatrixArgImm a
      (bPtr,ldb) <- submatrixArg_ b
      liftIO $ CBlas.addMatrix m n alpha aPtr lda beta bPtr ldb
