{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE TypeOperators #-}
module Numeric.BLAS.Matrix.RowMajor.Square (
   Matrix,
   size,
   identity,
   diagonal,
   takeDiagonal,
   ) where

import qualified Numeric.BLAS.Vector.SlicePrivate as VectorSlice
import Numeric.BLAS.Scalar (zero, one)

import qualified Numeric.BLAS.FFI.Generic as Blas
import qualified Numeric.Netlib.Utility as Call
import qualified Numeric.Netlib.Class as Class
import Numeric.BLAS.Private (pokeCInt)

import Foreign.ForeignPtr (withForeignPtr)
import Foreign.Storable (poke)

import Control.Monad.Trans.Cont (ContT(ContT), evalContT)
import Control.Monad.IO.Class (liftIO)

import qualified Data.Array.Comfort.Storable.Unchecked as Array
import qualified Data.Array.Comfort.Shape.SubSize as SubSize
import qualified Data.Array.Comfort.Shape as Shape
import Data.Array.Comfort.Storable.Unchecked (Array(Array))



{- $setup
>>> import Test.NumberType_.Numeric.BLAS.Vector (forVector, Number_, number_)
>>> import Test.Slice (ShapeInt, shapeInt)
>>> import qualified Numeric.BLAS.Matrix.RowMajor.Square as Square
>>> import qualified Numeric.BLAS.Vector as Vector
>>> import Numeric.BLAS.Vector (Vector)
>>> import qualified Test.QuickCheck as QC
>>> import Test.QuickCheck ((===))
>>>
>>> maxDim :: Int
>>> maxDim = 10
-}


type Matrix sh = Array (Shape.Square sh)
type Vector = Array

size :: Matrix sh a -> sh
size = Shape.squareSize . Array.shape

identity :: (Shape.C sh, Class.Floating a) => sh -> Matrix sh a
identity sh =
   Array.unsafeCreateWithAutoSizes (Shape.Square sh) $
      \(SubSize.Sub blockSize (Shape.Square (SubSize.Atom nint))) yPtr ->
   evalContT $ do
      nPtr <- Call.alloca
      xPtr <- Call.number zero
      incxPtr <- Call.cint 0
      incyPtr <- Call.cint 1
      liftIO $ do
         poke nPtr $ fromIntegral blockSize
         Blas.copy nPtr xPtr incxPtr yPtr incyPtr
         let n = fromIntegral nint
         poke nPtr n
         poke xPtr one
         poke incyPtr (n+1)
         Blas.copy nPtr xPtr incxPtr yPtr incyPtr

-- ToDo: use this in LAPACK.Square.Basic
{- |
prop> :{
   QC.forAll (fmap shapeInt $ QC.choose (0,maxDim)) $ \sh ->
   Square.identity sh
   ===
   Square.diagonal (Vector.one sh :: Vector ShapeInt Number_)
:}
-}
diagonal ::
   (VectorSlice.C v, Shape.C sh, Class.Floating a) => v sh a -> Matrix sh a
diagonal x =
   Array.unsafeCreateWithAutoSizes (Shape.Square $ VectorSlice.shape x) $
      \(SubSize.Sub blockSize (Shape.Square (SubSize.Atom nint))) yPtr ->
   evalContT $ do
      nPtr <- Call.alloca
      (xPtr,incxPtr) <- VectorSlice.sliceArg x
      zPtr <- Call.number zero
      incyPtr <- Call.cint 1
      inczPtr <- Call.cint 0
      liftIO $ do
         pokeCInt nPtr blockSize
         Blas.copy nPtr zPtr inczPtr yPtr incyPtr
         let n = fromIntegral nint
         poke nPtr n
         poke incyPtr (n+1)
         Blas.copy nPtr xPtr incxPtr yPtr incyPtr

{- |
prop> :{
   forVector number_ $ \xs ->
   xs === Square.takeDiagonal (Square.diagonal xs)
:}
-}
takeDiagonal :: (Shape.C sh, Class.Floating a) => Matrix sh a -> Vector sh a
takeDiagonal (Array (Shape.Square sh) x) =
   Array.unsafeCreateWithSize sh $ \n yPtr -> evalContT $ do
      nPtr <- Call.cint n
      xPtr <- ContT $ withForeignPtr x
      incxPtr <- Call.cint (n+1)
      incyPtr <- Call.cint 1
      liftIO $ Blas.copy nPtr xPtr incxPtr yPtr incyPtr
