{- ToDo:
We can also think about a Buffer type with a Shape type parameter.
It would support inplace update operations
and a final conversion to Arrays, but no sharing.
We would provide optimization rules
to fuse Vector and VectorSlice operations to Buffer operations.
E.g. VectorSlice.mac is a sliced copy and an inplace update, anyway.

If automatic fusion is not possible,
we could provide a dummy 'fuse' function,
that has the semantics of 'id'
but is a seed for fusion for the simplifier.
-}
{- |
ToDo: inplace raise, add, scale

Modifiers provide inplace updates.
This is still referentially transparent but does not support sharing.

E.g.

> let c = Chunk.conjugate $ Chunk.simple x
> in c <> c

This would compute @Chunk.conjugate@ twice.
-}
module Numeric.BLAS.Vector.Chunk (
   T,
   -- * finalizing chunk
   toVector,
   toStorableVector,
   toMutableVector,
   -- * chunk creators
   simple,
   reversed,
   singleton,
   replicate,
   zero,
   one,
   -- ToDo: submatrix :: BlockMatrix sh a -> Chunk a
   -- * chunk modifiers
   conjugate,
   ) where

import qualified Numeric.BLAS.Vector.SlicePrivate as VectorSlice
import qualified Numeric.BLAS.Scalar as Scalar
import Numeric.BLAS.Private (lacgv)

import qualified Numeric.BLAS.FFI.Generic as Blas
import qualified Numeric.Netlib.Utility as Call
import qualified Numeric.Netlib.Class as Class

import qualified Data.Array.Comfort.Storable.Mutable.Private as MutArray
import qualified Data.Array.Comfort.Storable.Unchecked as Array
import qualified Data.Array.Comfort.Shape as Shape
import qualified Data.StorableVector.Base as SVB
import Data.Array.Comfort.Storable.Unchecked (Array)

import qualified Foreign.Marshal.Utils as Marshal
import Foreign.Marshal.Array (advancePtr)
import Foreign.Ptr (Ptr)
import Foreign.Storable (Storable, poke)

import Data.Semigroup (Semigroup((<>)))
import Data.Monoid (Monoid(mempty,mappend))

import Control.Monad.Primitive (PrimMonad)
import Control.Monad.Trans.Cont (ContT(ContT), evalContT)
import Control.Monad.IO.Class (liftIO)

import Prelude hiding (replicate)


data T a =
   Cons {
      _size :: Int,
      _write :: Ptr a -> IO ()
   }

instance (Storable a) => Semigroup (T a) where
   Cons sizeX writeX <> Cons sizeY writeY =
      Cons (sizeX+sizeY) (\ptr -> writeX ptr >> writeY (advancePtr ptr sizeX))

instance (Storable a) => Monoid (T a) where
   mempty = Cons 0 (const $ return ())
   mappend = (<>)


type Vector = Array
type ShapeInt = Shape.ZeroBased Int

toVector :: (Storable a) => T a -> Vector ShapeInt a
toVector (Cons size_ write_) =
   Array.unsafeCreate (Shape.ZeroBased size_) write_

toStorableVector :: (Storable a) => T a -> SVB.Vector a
toStorableVector (Cons size_ write_) = SVB.unsafeCreate size_ write_

toMutableVector ::
   (PrimMonad m, Storable a) => T a -> m (MutArray.Array m ShapeInt a)
toMutableVector (Cons size_ write_) =
   MutArray.unsafeCreate (Shape.ZeroBased size_) write_


{-
ToDo:
define a new type class without the shape type parameter
could be used for StorableVector
-}
simple :: (VectorSlice.C v, Integral n, Class.Floating a) =>
   v (Shape.ZeroBased n) a -> T a
simple x =
   let n = Shape.size $ VectorSlice.shape x in
   Cons n $
      \yPtr -> evalContT $ do
         nPtr <- Call.cint n
         (xPtr, incxPtr) <- VectorSlice.sliceArg x
         incyPtr <- Call.cint 1
         liftIO $ Blas.copy nPtr xPtr incxPtr yPtr incyPtr

reversed :: (VectorSlice.C v, Integral n, Class.Floating a) =>
   v (Shape.ZeroBased n) a -> T a
reversed x =
   let n = Shape.size $ VectorSlice.shape x in
   Cons n $
      \yPtr -> evalContT $ do
         nPtr <- Call.cint n
         xPtr <- VectorSlice.startArg x
         incxPtr <- Call.cint $ negate $ VectorSlice.increment x
         incyPtr <- Call.cint 1
         liftIO $ Blas.copy nPtr xPtr incxPtr yPtr incyPtr


singleton :: (Storable a) => a -> T a
singleton x = Cons 1 (\ptr -> poke ptr x)

replicate :: (Integral n, Class.Floating a) => n -> a -> T a
replicate nn x =
   let n = fromIntegral nn in
   Cons n $
      \yPtr -> evalContT $ do
         nPtr <- Call.cint n
         xPtr <- ContT $ Marshal.with x
         incxPtr <- Call.cint 0
         incyPtr <- Call.cint 1
         liftIO $ Blas.copy nPtr xPtr incxPtr yPtr incyPtr

zero, one :: (Integral n, Class.Floating a) => n -> T a
zero n = replicate n Scalar.zero
one  n = replicate n Scalar.one



conjugate :: (Class.Floating a) => T a -> T a
conjugate (Cons n write_) =
   Cons n $ \ptr -> do
      write_ ptr
      evalContT $ do
         nPtr <- Call.cint n
         incxPtr <- Call.cint 1
         liftIO $ lacgv nPtr ptr incxPtr
